from __future__ import unicode_literals
from copy import copy
from fauxquests.response import Resp
from fauxquests.utils import URL, Registry
from requests.adapters import HTTPAdapter
from requests.compat import quote
from unittest import TestCase
import json
import six


class FauxAdapter(HTTPAdapter):
    """A subclass to HTTPAdapter that delivers pre-registered
    content in response to a request, or else raises an exception.
    """
    def __init__(self, url_pattern='%s'):
        """Initialize a FauxAdapter.

        This method is explicitly defined in order to **drop** support
        for the keyword arguments supported by HTTPAdapter (which no longer
        make sense in FauxAdapter's case).

        It adds support for a `url_pattern` argument, which if provided
        is used to interpolate URLs sent to `register` (but not `send`).
        """
        super(FauxAdapter, self).__init__()
        self._registry = Registry()
        self._url_pattern = url_pattern
        self.requests = []

    def clear(self):
        self._registry.clear()

    def register(self, url, response, status_code=200, headers=None, **kwargs):
        """Add the given URL to the registry, and assign the response
        to it.

        If keyword arguments are provided, convert them into a query
        string for the URL. Query string keys and values must be present
        in order for a URL to match (but callers can send a superset and
        still match).
        """
        # Interpolate the provided URL pattern.
        # (This defaults to '%s', which will effectively be a no-op.)
        url = self._url_pattern % url

        # Create a URL object from the URL and keyword arguments.
        url = URL(url, **kwargs)

        # If the response is a string object, convert it to a bytes object.
        if isinstance(response, six.text_type):
            response = response.encode('utf8')

        # Add this response to the registry.
        headers = headers or {}
        self._registry[url] = Resp(response, status_code, headers)

    def register_json(self, url, response, status_code=200,
                      headers=None, **kwargs):
        """Dump the response to JSON, add the appropriate headers to match,
        and register the URL as normal.
        """
        # Properly state that JSON will be sent down.
        headers = headers or {}
        headers['Content-Type'] = 'application/json'

        # Dump the response to a JSON string.
        response = json.dumps(response)

        # Perform the registration.
        return self.register(url, response, status_code=status_code,
                             headers=headers, **kwargs)

    def send(self, request, stream=False, **kwargs):
        """Return the appropriate pre-registered response when a request
        is made.

        If the URL is not matched, the registry raises UnregisteredURL.
        """
        # Get the pre-registered response from our registry.
        url = URL(request.url)
        response = copy(self._registry[url])

        # Use requests to build the response object.
        r = self.build_response(request, response)

        # If this isn't a streaming request, pre-load the content.
        #
        # This is taken from `requests_testadapter`; I'm blindly assuming
        # he has a good reason.
        if not stream:
            r.content

        # Track information about this call, so that asserts can be made
        # against what was done on this object.
        self.requests.append(request)

        # Return the response object
        return r



from __future__ import unicode_literals
from collections import namedtuple
from fauxquests.adapter import FauxAdapter
from fauxquests.compat import mock
from requests.compat import OrderedDict
from requests.sessions import Session
from sdict import AlphaSortedDict


class FauxServer(Session):
    """A class that can register certain endpoints to have false
    responses returned.
    """
    def __init__(self, adapter_class=FauxAdapter, url_pattern='%s'):
        """Create a new Fauxquests instance, which knows how to
        mock out requests.session.Session and insert itself.

        If a `url_pattern` is provided, then all URLs registered
        are interpolated through the `url_pattern`.
        """
        # Initialize this object.
        super(FauxServer, self).__init__()
        self.patcher = mock.patch('requests.sessions.Session',
                                  return_value=self)
        self.adapters = OrderedDict()

        # Write settings to this object.
        self.adapter_class = adapter_class
        self.url_pattern = url_pattern

        # Save a list of registrations to apply to any FauxAdapter
        # that this FauxServer creates.
        self.registrations = {}

    def __enter__(self):
        """Mock out `requests.session.Session`, replacing it with this
        object.
        """
        return self.start()

    def __exit__(self, type, value, traceback):
        return self.stop()

    def register(self, url, response, status_code=200, headers=None, **kwargs):
        """Register a given URL and response with this FauxServer.

        Internally, this object's context manager creates and returns a
        FauxAdapters, so regisrations within a context manager go away
        when the context manager is exited.

        This method, however, is run before the context manager is applied,
        and applies universally to all adapters this object creates.
        """
        self.registrations[url] = Registration('', response, status_code,
                                              headers, kwargs)

    def register_json(self, url, response, status_code=200,
                      headers=None, **kwargs):
        """Register a given URL and response with this FauxServer.

        Internally, this object's context manager creates and returns a
        FauxAdapters, so regisrations within a context manager go away
        when the context manager is exited.

        This method, however, is run before the context manager is applied,
        and applies universally to all adapters this object creates.
        """
        self.registrations[url] = Registration('json', response, status_code,
                                               headers, kwargs)

    def start(self):
        """Institute the patching process, meaining requests sent to
        requests (how meta) are caught and handled by our adapter instead.
        """
        # Mount the Fauxquests adapter, which handles delivery of
        # responses based on the provided URL.
        adapter = self.adapter_class(url_pattern=self.url_pattern)
        self.mount('https://', adapter)
        self.mount('http://', adapter)

        # Iterate over any registrations that are saved as part of this
        # FauxServer object and register them to the Adapter.
        for url, reg in self.registrations.items():
            # Is this a plain registration or a JSON registration?
            method_name = 'register'
            if reg.type:
                method_name += '_' + reg.type

            # Forward the registration to the adapter.
            getattr(adapter, method_name)(url, reg.response, reg.status_code,
                                          reg.headers, **reg.kwargs)

        # Start the patcher.
        self.patcher.start()

        # Return the adapter object, which can accept registred
        # URLs with responses
        return adapter

    def stop(self):
        """Undo the patching process set up in `self.start`, and also
        set this object back to having no adapters.
        """
        self.patcher.stop()
        self.adapters = OrderedDict()


Registration = namedtuple('Registration', ['type', 'response', 'status_code',
                                           'headers', 'kwargs'])



# -*- coding: utf-8 -*-
# This response class is copied, with only slight modifications, from
# Łukasz Langa's `requests_testadapter` package (which was a springboard
# for this project).
#
# Therefore, the following copyright / license notice applies to the code
# in this file:
#
# Copyright (C) 2013 by Łukasz Langa
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from six import BytesIO

import requests.status_codes


class Resp(BytesIO):
    def __init__(self, stream, status=200, headers=None):
        self.status = status
        self.headers = headers or {}
        self.reason = requests.status_codes._codes.get(
            status, ['']
        )[0].upper().replace('_', ' ')
        BytesIO.__init__(self, stream)

    @property
    def _original_response(self):
        return self

    @property
    def msg(self):
        return self

    def read(self, chunk_size, **kwargs):
        return BytesIO.read(self, chunk_size)

    def info(self):
        return self

    def get_all(self, name, default):
        result = self.headers.get(name)
        if not result:
            return default
        return [result]

    def getheaders(self, name):
        return self.get_all(name, [])

    def release_conn(self):
        self.close()



#!/usr/bin/env python
from distutils.core import setup
from os.path import dirname, realpath
from os import sep
from setuptools import find_packages
from setuptools.command.test import test as TestCommand
import sys


pip_requirements = realpath(dirname(__file__)) + sep + 'requirements.txt'


class Tox(TestCommand):
    """The test command should install and then run tox.

    Based on http://tox.readthedocs.org/en/latest/example/basic.html
    """
    def finalize_options(self):
        super().finalize_options()
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import tox  # Import here, because outside eggs aren't loaded.
        sys.exit(tox.cmdline(self.test_args))


setup(
    # Basic metadata
    name='fauxquests',
    version=open('VERSION').read().strip(),
    author='Luke Sneeringer',
    author_email='luke@sneeringer.com',
    url='https://github.com/lukesneeringer/fauxquests',

    # Additional information
    description=' '.join((
        'Mock HTTP requests sent with the requests package.',
        'For Python 2 or Python 3.',
    )),
    license='New BSD',

    # What does this require?
    install_requires=open(pip_requirements, 'r').read().strip().split('\n'),

    # How to do the install
    provides=[
        'fauxquests',
    ],
    packages=[i for i in find_packages() if i.startswith('fauxquests')],

    # How to do the tests
    tests_require=['tox'],
    cmdclass={'test': Tox },

    # Data files
    package_data={
        'fauxquests': ['VERSION'],
    },

    # PyPI metadata
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Topic :: Software Development',
    ],
)



from fauxquests.exceptions import UnregisteredURL
from fauxquests.messages import NOT_FOUND
from requests.compat import quote
from sdict import AlphaSortedDict
from six.moves.urllib.parse import parse_qs
import six


class Registry(dict):
    """A dict subclass that understands how to handle URLs with
    keyword argument subsets for lookups.
    """
    def __getitem__(self, key):
        """Return a response corresponding to the given key."""

        # Sanity check: This might be easy.
        # If the URL is an exact match via. the regular `dict` lookup,
        # then just return that.
        no = object()
        super_answer = super(Registry, self).get(key, no)
        if super_answer != no:
            return super_answer

        # Convert the key to a URL object if it is not already one.
        if isinstance(key, (six.text_type, six.binary_type)):
            key = URL(key)

        # It's still possible we have a match.
        # Iterate over each key and see if our key is an exact superset of
        # the given key.  If any are, return its value immediately.
        for contender in self.keys():
            if key.is_exact_superset_of(contender):
                return self[contender]

        # Nope, no match. :(
        # Now, we need to raise an exception.
        # First, Get a list of the registered URLs in a nice format.
        registered_urls = '\n    - '.join(
            [six.text_type(i) for i in six.iterkeys(self)],
        )
        if not registered_urls:
            registered_urls = '(None)'

        # Raise an exception.
        # UnregisteredURL inherits from AssertionError rather than KeyError,
        # which should give more attractive failures in tests.
        raise UnregisteredURL(NOT_FOUND.format(
            registered_urls=registered_urls,
            request_url=key,
        ))


class URL(object):
    """A class for holding a URL, with convenience methods
    for poking at its query string separately from the main URI.
    """
    def __init__(self, url, **kwargs):
        # Does the URL string have a query string in it?
        # If so, parse that out into keyword arguments.
        if '?' in url:
            qs = parse_qs(url[url.index('?') + 1:])
            for k, v in qs.items():
                if len(v) == 1:
                    v = v[0]
                else:
                    v = set(v)
                kwargs.setdefault(k, v)
            url = url[:url.index('?')]

        # Store the base URL and the keyword arguments.
        self.uri = url
        self.qs = AlphaSortedDict(kwargs)

    def __eq__(self, other):
        # If this is a string of some kind, convert it to a URL.
        if isinstance(other, (six.text_type, six.binary_type)):
            other = self.__class__(other)

        # Type check!
        if not hasattr(other, 'qs') or not hasattr(other, 'uri'):
            raise TypeError('Cannot compare URL with %s.' %
                            other.__class__.__name__)

        # Return whether the URI and QS are the same.
        return self.uri == other.uri and self.qs == other.qs

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        """Return the full URL, as a string."""

        # Start with the basic URI.
        answer = self.uri

        # Append any keyword arguments to the query-string
        if self.qs:
            qs_list = []
            for key, values in self.qs.items():
                # Convert `values` to a list if it's not one already.
                # This way, we seemlessly address situations where the
                # same key is used multiple times.
                if isinstance(values, (list, tuple, set)):
                    values = list(values)
                    values.sort()
                else:
                    values = [values]

                # Iterate over the values and add each to the answer.
                for value in values:
                    qs_list.append('%s=%s' % (key, quote(str(value))))

            # Append the query string to the answer.
            answer += '?' + '&'.join(qs_list)

        # Return the answer.
        return answer

    def is_exact_superset_of(self, other):
        """Return True if this URL is an exact superset of the URL provided
        by `other`.

        Definition of exact superset is:
          - `self.uri` must exactly match `other.uri`.
          - Every key in `other.qs` must be present in `self.qs` with
            the same value.
        """
        # If this is a string of some kind, convert it to a URL.
        if isinstance(other, (six.text_type, six.binary_type)):
            other = self.__class__(other)

        # Type check!
        if not hasattr(other, 'qs') or not hasattr(other, 'uri'):
            raise TypeError('Cannot compare URL with %s.' %
                            other.__class__.__name__)

        # If the URIs do not exactly match, this is not an exact superset.
        if self.uri != other.uri:
            return False

        # Ensure that every key in `other.qs` exists in `self.qs` with
        # the same value; if any do not, immediately return False.
        for key in other.qs.keys():
            # If the key doesn't exist, return False.
            if key not in self.qs:
                return False

            # If the value does not match, return False.
            if self.qs[key] != other.qs[key]:
                return False

        # Okay, we must have an exact superset!
        return True



from __future__ import unicode_literals

# Import mock from the unittest module in the stdlib if it's there
# (Python 3.3), otherwise import the pip package, which is required
# in Python 2.
try:
    from unittest import mock
except ImportError:
    import mock

# Import unittest2 if we have it (required on Python 2.6),
# otherwise just use the stdlib unittest.
try:
    from unittest2 import unittest
except ImportError:
    import unittest
